// © 2023, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

#pragma once

#include "src/math/Vec4f.h"

/**
 * Defines a ray with an origin position and a direction vector.
 */
struct Ray
{
    /** Origin of this ray */
    Vec4f origin;

    /** The direction of this ray */
    Vec4f direction;

public:
    /**
     * Initialize an invalid ray per default.
     */
    Ray();

    /**
     * Initialize the ray with the given origin point and the given direction vector.
     *
     * @param pOrigin       Origin point of the ray.
     * @param pDirection    Direction vector of the ray.
     */
    Ray(Vec4f pOrigin, Vec4f pDirection);

    /**
     * Reflects this ray at the given hitPoint with the given surface normal
     * and returns it.
     *
     * @param hitPoint          Intersection point on which the ray is reflected.
     * @param surfaceNormal     Normal of the surface at the intersection point.
     *
     * @return The reflected ray.
     */
    Ray reflected(Vec4f hitPoint, Vec4f surfaceNormal);

    /**
     * Refracts this ray at the given hitPoint with the given surface normal
     * and returns it.
     *
     * If there is no result, the direction of the returned ray is (0,0,0).
     *
     * @param hitPoint          Intersection point on which the ray is reflected.
     * @param surfaceNormal     Normal of the surface at the intersection point.
     * @param refractiveIndex   Refractive index.
     *
     * @return  The refracted ray. If there is no valid refracted ray, the
     *          direction of the returned ray is (0,0,0).
     */
    Ray refracted(Vec4f hitPoint, Vec4f surfaceNormal, float refractiveIndex);

    /**
     * Returns the point: p = ray.origin + ray.direction * t
     */
    Vec4f evaluate(float t) const;

    /**
     * Returns whether this ray is valid. This is the case if the direction
     * vector has not the length 0 and has no nan values.
     */
    bool isValid();
};
