#include "Raytracer.h"

#include <vector>
#include <algorithm>
#include <random>
#include <ranges>

#include "src/shader/Shader.h"

#include "src/scene/SurfaceList.h"

Raytracer::Raytracer(int& renderResolution, int& previewResolution)
    : renderResolution(renderResolution)
    , previewResolution(previewResolution)
{
    // Create raytracing thread which calls the render function if wanted:
    raytraceThread = std::thread([this](){
        while(!waitForTerminate){
			if (!displayPreviewImage) {
				isRendering = false;
				raytraceSemaphore.acquire();
				isRendering = true;
			}

			bool renderInPreviewMode = previewMode;

			resizeImagesIfNecessary();

			if (waitForTerminate)
				break;

			render(renderInPreviewMode);
        }
    });

    raytraceSemaphore.release();
}

Raytracer::~Raytracer(){
    waitForTerminate = true;

    raytraceSemaphore.release();
    raytraceThread.join();
}

float Raytracer::getProgress(){
    return progress;
}

float Raytracer::getRunningTime(){
    return runningTime;
}

void Raytracer::initPixelIDMap(std::vector<unsigned int>& pixelIDMap, unsigned int size){
    pixelIDMap = std::vector<unsigned int>(size);
    for(unsigned int i = 0; i < size; ++i){
        pixelIDMap[i] = i;
    }

    std::random_device rd;
    std::mt19937 gen {rd()};
    std::ranges::shuffle(pixelIDMap, gen);
}

void Raytracer::resizeImagesIfNecessary(){
    int renderSize = renderResolution * renderResolution;
    if(int(renderedImage.size()) != renderSize){
        initPixelIDMap(renderPixelIDMap, renderSize);
        renderedImage = std::vector<Vec4f>(renderSize);
    }

    int previewSize = previewResolution * previewResolution;
    if(int(previewImage.size()) != previewSize){
        initPixelIDMap(previewPixelIDMap, previewSize);
        previewImage = std::vector<Vec4f>(previewSize);
    }
}

void Raytracer::copyPreviewToFull(){
    #pragma omp parallel num_threads(16)
    {
        float scaleFactor = float(previewResolution) / renderResolution;

        #pragma omp for
        for(int x = 0; x < int(renderResolution); ++x){
            for(int y = 0; y < int(renderResolution); ++y){
                int pX = int(x * scaleFactor);
                int pY = int(y * scaleFactor);

                renderedImage[y * renderResolution + x] = previewImage[pY * previewResolution + pX];
            }
        }
    }
}

void Raytracer::render(bool renderInPreviewMode){
    // Copy local variables:
    int resolution = renderResolution;
    std::vector<Vec4f>* image = &renderedImage;
    std::vector<unsigned int>* pixelIDMap = &renderPixelIDMap;

    shouldInterrupt = false;
    // If preview mode, we have to use other variables:
    if(renderInPreviewMode){
        resolution = previewResolution;
        image = &previewImage;
        pixelIDMap = &previewPixelIDMap;
        progress = 0;
    } else {
        copyPreviewToFull();
        displayPreviewImage = false;
    }

    scene.surfaces.setAccelerator( accelerator );
    auto start_time = std::chrono::high_resolution_clock::now();

    // Generate a ray for each pixel and raytrace:
    for(int x = 0; x < resolution; ++x){

        #pragma omp parallel num_threads(16)
        {
            #pragma omp for
            for(int y = 0; y < resolution; ++y){
                // Pixel ID:
                unsigned int id = pixelIDMap->at(y * resolution + x);
                unsigned int rX = id % resolution;
                unsigned int rY = id / resolution;

                // The ray which should be traced for this pixel:
                Ray ray = scene.camera.generateRay(rX, rY, resolution, resolution);

                // Store the color into the image array:
                image->at(id) = traceRay(ray, maxRayDepth);
            }

        }

        if(!renderInPreviewMode && previewMode){
            break;
        }

        if(shouldInterrupt){
            shouldInterrupt = false;
            break;
        }

        // Update progress bar:
        if(!renderInPreviewMode) {
            progress = x / float(resolution);
            auto intermediate_time = std::chrono::high_resolution_clock::now();
            runningTime = std::chrono::duration<float>(intermediate_time-start_time).count();
        }
    }

    // Avoid double rendering in full resolution:
    raytraceSemaphore.setCount(0);
}


void Raytracer::setPreviewMode(bool mode) {
    previewMode = mode;

    // If preview mode is activated, ensure that rendering gets reactivated:
    if(mode){
        if(raytraceSemaphore.getCount() <= 0)
            raytraceSemaphore.release();

        displayPreviewImage = true;
    }
}

void Raytracer::setScene(Scene pScene){
    scene = pScene;

    // Trigger repainting:
    setPreviewMode(true);
}

Vec4f Raytracer::traceRay(Ray ray, unsigned int depth){
    // If maximum depth is reached, just return black:
    if (depth == 0)
        return Vec4f();

    // Perform an intersection test agaist the surface (all surfaces):
    Intersection intersection = scene.surfaces.intersect(ray, 0.001f, 20.f);

    // If there is an intersection, shoot secondary rays and shade:
    if (intersection.hasIntersection) {
        Vec4f color = shade(intersection, ray);

        Shader* shader = intersection.shader;
        Vec4f& hitPoint = intersection.hitPoint;
        Vec4f& normal = intersection.normal;

        // Trace reflected ray recursively (only if required):
        if(shader->hasReflective()){
            Ray reflectedRay = ray.reflected(hitPoint, normal);
            color = color + shader->getReflectiveColor() * traceRay(reflectedRay, depth - 1);
        }

        // Trace refracted ray recursively (only if required):
        if(shader->hasRefractive()){
            Ray refractedRay = ray.refracted(hitPoint, normal, shader->getRefractiveIndex());

            // Only shoot secondary ray if a valid refracted ray exists:
            if(refractedRay.isValid()){
                color = color + shader->getRefractiveColor() * traceRay(refractedRay, depth - 1);
            }
        }

        return color;
    }

    // Otherwise, return background color:
    return backgroundColor;
}

Vec4f Raytracer::shade(Intersection& intersection, Ray& ray){
    Vec4f color = intersection.shader->getAmbientColor();

    // Iterate over all lights:
    for(PointLight& light : scene.lights){
        Ray shadowRay(Vec4f(intersection.hitPoint), light.getPosition() - intersection.hitPoint);

        Intersection shadowTest = scene.surfaces.intersect(shadowRay, 0.001f, 0.999f);
        if(!shadowTest.hasIntersection){
            color += intersection.shader->shade(light, intersection, ray) + intersection.shader->shadeAmbient(light.getColor());
        }
    }

    return color;
}
