#include "SceneLoader.h"

#include "src/scene/Scene.h"

#include "src/scene/surfaces/Checkerboard.h"
#include "src/scene/surfaces/Plane.h"
#include "src/scene/surfaces/Sphere.h"
#include "src/scene/Mesh.h"

#include <fstream>
#include <memory>

using json = nlohmann::json;

Scene SceneLoader::loadScene(std::string path){
    using namespace nlohmann;

    Scene scene;

    std::ifstream file(path);
    json data;
    try {
        data = json::parse(file);

        Vec4f cameraTarget(data["camera"]["target"].get<std::vector<float>>());
        float cameraDistance = data["camera"]["distance"].get<float>();
        float focalLength = data["camera"]["focalLength"].get<float>();

        float xRotation = data["camera"].contains("xRotation") ? data["camera"]["xRotation"].get<float>() : 0;
        float yRotation = data["camera"].contains("yRotation") ? data["camera"]["yRotation"].get<float>() : 0;

        scene.camera = Camera(cameraTarget, cameraDistance, focalLength, xRotation, yRotation);

        std::vector<Shader> shaders;
        // Materials
        for (const auto& item : data["materials"].items())
        {
            Shader shader;
            shader.ambientColor = Vec4f(item.value()["ambient"].get<std::vector<float>>());
            shader.diffuseColor = Vec4f(item.value()["diffuse"].get<std::vector<float>>());
            shader.reflectiveColor = Vec4f(item.value()["reflective"].get<std::vector<float>>());
            shader.refractiveColor = Vec4f(item.value()["refractive"].get<std::vector<float>>());
            shader.refractiveIndex = item.value()["refractiveIndex"].get<float>();
            shader.shininess = item.value()["shininess"].get<float>();
            shaders.push_back(shader);
        }

        // Surfaces
        for (const auto& item : data["surfaces"].items())
        {
            if(item.value()["type"] == "Plane"){
                scene.surfaces.addSurface(std::make_shared<Plane>(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["normal"].get<std::vector<float>>()).normalized(),
                    shaders[item.value()["material"].get<int>()]
                    ));
            } if(item.value()["type"] == "Sphere"){
                scene.surfaces.addSurface(std::make_shared<Sphere>(
                    Vec4f(item.value()["center"].get<std::vector<float>>()),
                    item.value()["radius"].get<float>(),
                    shaders[item.value()["material"].get<int>()]
                    ));
            } else if(item.value()["type"] == "Checkerboard"){
                scene.surfaces.addSurface(std::make_shared<Checkerboard>(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["x"].get<std::vector<float>>()),
                    Vec4f(item.value()["y"].get<std::vector<float>>()),
                    shaders[item.value()["material"].get<int>()],
                    shaders[item.value()["material2"].get<int>()]
                ));
            } else if(item.value()["type"] == "Mesh") {
                auto mesh = Mesh(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["scale"].get<std::vector<float>>()),
                    shaders[item.value()["material"].get<int>()],
                    item.value()["fileName"]
                );
                scene.surfaces.addSurfaces(mesh.getTriangles());
            } else if(item.value()["type"] == "MeshSet") {
                auto counts = item.value()["counts"].get<std::vector<int>>();
                auto positionStart = Vec4f(item.value()["positionStart"].get<std::vector<float>>());
                auto positionEnd = Vec4f(item.value()["positionEnd"].get<std::vector<float>>());
                auto scale = Vec4f(item.value()["scale"].get<std::vector<float>>());
                auto shader = shaders[item.value()["material"].get<int>()];
                std::string fileName = item.value()["fileName"];
                for (float x=1; x<=counts[0]; ++x) {
                    float interp_x = static_cast<float>(x-.5)/static_cast<float>(counts[0]);
                    auto pos_x = positionStart.x*(1-interp_x)+positionEnd.x*interp_x;
                    for (int y=1; y<=counts[1]; ++y) {
                        float interp_y = static_cast<float>(y-.5)/static_cast<float>(counts[1]);
                        auto pos_y = positionStart.y*(1-interp_y)+positionEnd.y*interp_y;
                        for (int z=1; z<=counts[2]; ++z) {
                            float interp_z = static_cast<float>(z-.5)/static_cast<float>(counts[2]);
                            auto pos_z = positionStart.z*(1-interp_z)+positionEnd.z*interp_z;
                            auto position = Vec4f(pos_x, pos_y, pos_z, 1);
                            auto mesh = Mesh(position, scale, shader, fileName);
                            scene.surfaces.addSurfaces(mesh.getTriangles());
                        }
                    }
                }
            }
        }

        // Lights:
        std::vector<PointLight> lights;
        for (const auto& item : data["lights"].items())
        {
            if(item.value()["type"] == "PointLight"){
                lights.push_back(PointLight(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["color"].get<std::vector<float>>())
                ));
            }
        }
        scene.lights = lights;
    } catch(json::exception& ex){
        std::cout << ex.what() << std::endl;
    }

    return scene;
}
